"""JSON Web Algorithms.

https://tools.ietf.org/html/draft-ietf-jose-json-web-algorithms-40

"""

import abc
import logging
from collections.abc import Hashable
from typing import Any, Callable, Dict

import cryptography.exceptions
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, hmac
from cryptography.hazmat.primitives.asymmetric import ec, padding, rsa
from cryptography.hazmat.primitives.asymmetric.utils import (
    decode_dss_signature,
    encode_dss_signature,
)
from cryptography.hazmat.primitives.hashes import HashAlgorithm

from josepy import errors, interfaces, jwk

logger = logging.getLogger(__name__)


class JWA(interfaces.JSONDeSerializable):
    # for some reason disable=abstract-method has to be on the line
    # above...
    """JSON Web Algorithm."""


class JWASignature(JWA, Hashable):
    """Base class for JSON Web Signature Algorithms."""

    SIGNATURES: Dict[str, "JWASignature"] = {}
    kty: Any

    def __init__(self, name: str) -> None:
        self.name = name

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, JWASignature):
            return NotImplemented
        return self.name == other.name

    def __hash__(self) -> int:
        return hash((self.__class__, self.name))

    @classmethod
    def register(cls, signature_cls: "JWASignature") -> "JWASignature":
        """Register class for JSON deserialization."""
        cls.SIGNATURES[signature_cls.name] = signature_cls
        return signature_cls

    def to_partial_json(self) -> Any:
        return self.name

    @classmethod
    def from_json(cls, jobj: Any) -> "JWASignature":
        return cls.SIGNATURES[jobj]

    @abc.abstractmethod
    def sign(self, key: Any, msg: bytes) -> bytes:  # pragma: no cover
        """Sign the ``msg`` using ``key``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def verify(self, key: Any, msg: bytes, sig: bytes) -> bool:  # pragma: no cover
        """Verify the ``msg`` and ``sig`` using ``key``."""
        raise NotImplementedError()

    def __repr__(self) -> str:
        return self.name


class _JWAHS(JWASignature):
    kty = jwk.JWKOct

    def __init__(self, name: str, hash_: Callable[[], HashAlgorithm]):
        super().__init__(name)
        self.hash = hash_()

    def sign(self, key: bytes, msg: bytes) -> bytes:
        signer = hmac.HMAC(key, self.hash, backend=default_backend())
        signer.update(msg)
        return signer.finalize()

    def verify(self, key: bytes, msg: bytes, sig: bytes) -> bool:
        verifier = hmac.HMAC(key, self.hash, backend=default_backend())
        verifier.update(msg)
        try:
            verifier.verify(sig)
        except cryptography.exceptions.InvalidSignature as error:
            logger.debug(error, exc_info=True)
            return False
        else:
            return True


class _JWARSA:
    kty = jwk.JWKRSA
    padding: Any = NotImplemented
    hash: HashAlgorithm = NotImplemented

    def sign(self, key: rsa.RSAPrivateKey, msg: bytes) -> bytes:
        """Sign the ``msg`` using ``key``."""
        try:
            return key.sign(msg, self.padding, self.hash)
        except AttributeError as error:
            logger.debug(error, exc_info=True)
            raise errors.Error("Public key cannot be used for signing")
        except ValueError as error:  # digest too large
            logger.debug(error, exc_info=True)
            raise errors.Error(str(error))

    def verify(self, key: rsa.RSAPublicKey, msg: bytes, sig: bytes) -> bool:
        """Verify the ``msg` and ``sig`` using ``key``."""
        try:
            key.verify(sig, msg, self.padding, self.hash)
        except cryptography.exceptions.InvalidSignature as error:
            logger.debug(error, exc_info=True)
            return False
        else:
            return True


class _JWARS(_JWARSA, JWASignature):
    def __init__(self, name: str, hash_: Callable[[], HashAlgorithm]) -> None:
        super().__init__(name)
        self.padding = padding.PKCS1v15()
        self.hash = hash_()


class _JWAPS(_JWARSA, JWASignature):
    def __init__(self, name: str, hash_: Callable[[], HashAlgorithm]) -> None:
        super().__init__(name)
        self.padding = padding.PSS(mgf=padding.MGF1(hash_()), salt_length=padding.PSS.MAX_LENGTH)
        self.hash = hash_()


class _JWAEC(JWASignature):
    kty = jwk.JWKEC

    def __init__(self, name: str, hash_: Callable[[], HashAlgorithm]):
        super().__init__(name)
        self.hash = hash_()

    def sign(self, key: ec.EllipticCurvePrivateKey, msg: bytes) -> bytes:
        """Sign the ``msg`` using ``key``."""
        sig = self._sign(key, msg)
        dr, ds = decode_dss_signature(sig)
        length = jwk.JWKEC.expected_length_for_curve(key.curve)
        return dr.to_bytes(length=length, byteorder="big") + ds.to_bytes(
            length=length, byteorder="big"
        )

    def _sign(self, key: ec.EllipticCurvePrivateKey, msg: bytes) -> bytes:
        try:
            return key.sign(msg, ec.ECDSA(self.hash))
        except AttributeError as error:
            logger.debug(error, exc_info=True)
            raise errors.Error("Public key cannot be used for signing")
        except ValueError as error:  # digest too large
            logger.debug(error, exc_info=True)
            raise errors.Error(str(error))

    def verify(self, key: ec.EllipticCurvePublicKey, msg: bytes, sig: bytes) -> bool:
        """Verify the ``msg` and ``sig`` using ``key``."""
        rlen = jwk.JWKEC.expected_length_for_curve(key.curve)
        if len(sig) != 2 * rlen:
            # Format error - rfc7518 - 3.4 … MUST NOT be shortened to omit any leading zero octets
            return False
        asn1sig = encode_dss_signature(
            int.from_bytes(sig[0:rlen], byteorder="big"),
            int.from_bytes(sig[rlen:], byteorder="big"),
        )
        return self._verify(key, msg, asn1sig)

    def _verify(self, key: ec.EllipticCurvePublicKey, msg: bytes, asn1sig: bytes) -> bool:
        try:
            key.verify(asn1sig, msg, ec.ECDSA(self.hash))
        except cryptography.exceptions.InvalidSignature as error:
            logger.debug(error, exc_info=True)
            return False
        else:
            return True


#: HMAC using SHA-256
HS256 = JWASignature.register(_JWAHS("HS256", hashes.SHA256))
#: HMAC using SHA-384
HS384 = JWASignature.register(_JWAHS("HS384", hashes.SHA384))
#: HMAC using SHA-512
HS512 = JWASignature.register(_JWAHS("HS512", hashes.SHA512))

#: RSASSA-PKCS-v1_5 using SHA-256
RS256 = JWASignature.register(_JWARS("RS256", hashes.SHA256))
#: RSASSA-PKCS-v1_5 using SHA-384
RS384 = JWASignature.register(_JWARS("RS384", hashes.SHA384))
#: RSASSA-PKCS-v1_5 using SHA-512
RS512 = JWASignature.register(_JWARS("RS512", hashes.SHA512))

#: RSASSA-PSS using SHA-256 and MGF1 with SHA-256
PS256 = JWASignature.register(_JWAPS("PS256", hashes.SHA256))
#: RSASSA-PSS using SHA-384 and MGF1 with SHA-384
PS384 = JWASignature.register(_JWAPS("PS384", hashes.SHA384))
#: RSASSA-PSS using SHA-512 and MGF1 with SHA-512
PS512 = JWASignature.register(_JWAPS("PS512", hashes.SHA512))

#: ECDSA using P-256 and SHA-256
ES256 = JWASignature.register(_JWAEC("ES256", hashes.SHA256))
#: ECDSA using P-384 and SHA-384
ES384 = JWASignature.register(_JWAEC("ES384", hashes.SHA384))
#: ECDSA using P-521 and SHA-512
ES512 = JWASignature.register(_JWAEC("ES512", hashes.SHA512))



"""Shim class to not have to depend on typing module in prod."""

# mypy: ignore-errors
import sys
import warnings
from typing import Any

warnings.warn(
    "josepy.magic_typing is deprecated and will be removed in a future release.", DeprecationWarning
)


class TypingClass:
    """Ignore import errors by getting anything"""

    def __getattr__(self, name: str) -> Any:
        return None


try:
    # mypy doesn't respect modifying sys.modules
    from typing import *  # noqa: F401,F403
except ImportError:
    sys.modules[__name__] = TypingClass()



"""JOSE errors."""

from typing import Any


class Error(Exception):
    """Generic JOSE Error."""


class DeserializationError(Error):
    """JSON deserialization error."""

    def __str__(self) -> str:
        return "Deserialization error: {0}".format(super().__str__())


class SerializationError(Error):
    """JSON serialization error."""


class UnrecognizedTypeError(DeserializationError):
    """Unrecognized type error.

    :ivar str typ: The unrecognized type of the JSON object.
    :ivar jobj: Full JSON object.

    """

    def __init__(self, typ: str, jobj: Any) -> None:
        self.typ = typ
        self.jobj = jobj
        super().__init__(str(self))

    def __str__(self) -> str:
        return "{0} was not recognized, full message: {1}".format(self.typ, self.jobj)



"""JOSE utilities."""

import abc
import sys
import warnings
from collections.abc import Hashable, Mapping
from types import ModuleType
from typing import Any, Callable, Iterator, List, Tuple, TypeVar, Union, cast

from cryptography.hazmat.primitives.asymmetric import ec, rsa


# Deprecated. Please use built-in decorators @classmethod and abc.abstractmethod together instead.
def abstractclassmethod(func: Callable) -> classmethod:
    return classmethod(abc.abstractmethod(func))


class ComparableKey:
    """Comparable wrapper for ``cryptography`` keys.

    See https://github.com/pyca/cryptography/issues/2122.

    """

    __hash__: Callable[[], int] = NotImplemented

    def __init__(
        self,
        wrapped: Union[
            rsa.RSAPrivateKeyWithSerialization,
            rsa.RSAPublicKeyWithSerialization,
            ec.EllipticCurvePrivateKeyWithSerialization,
            ec.EllipticCurvePublicKeyWithSerialization,
        ],
    ):
        self._wrapped = wrapped

    def __getattr__(self, name: str) -> Any:
        return getattr(self._wrapped, name)

    def __eq__(self, other: Any) -> bool:
        if (
            not isinstance(other, self.__class__)
            or self._wrapped.__class__ is not other._wrapped.__class__
        ):
            return NotImplemented
        elif hasattr(self._wrapped, "private_numbers"):
            return self.private_numbers() == other.private_numbers()
        elif hasattr(self._wrapped, "public_numbers"):
            return self.public_numbers() == other.public_numbers()
        else:
            return NotImplemented

    def __repr__(self) -> str:
        return "<{0}({1!r})>".format(self.__class__.__name__, self._wrapped)

    def public_key(self) -> "ComparableKey":
        """Get wrapped public key."""
        if isinstance(
            self._wrapped,
            (rsa.RSAPublicKeyWithSerialization, ec.EllipticCurvePublicKeyWithSerialization),
        ):
            return self

        return self.__class__(self._wrapped.public_key())


class ComparableRSAKey(ComparableKey):
    """Wrapper for ``cryptography`` RSA keys.

    Wraps around:

    - :class:`~cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey`
    - :class:`~cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey`

    """

    def __hash__(self) -> int:
        # public_numbers() hasn't got stable hash!
        # https://github.com/pyca/cryptography/issues/2143
        if isinstance(self._wrapped, rsa.RSAPrivateKeyWithSerialization):
            priv = self.private_numbers()
            pub = priv.public_numbers
            return hash(
                (self.__class__, priv.p, priv.q, priv.dmp1, priv.dmq1, priv.iqmp, pub.n, pub.e)
            )
        elif isinstance(self._wrapped, rsa.RSAPublicKeyWithSerialization):
            pub = self.public_numbers()
            return hash((self.__class__, pub.n, pub.e))

        raise NotImplementedError()


class ComparableECKey(ComparableKey):
    """Wrapper for ``cryptography`` EC keys.
    Wraps around:
    - :class:`~cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey`
    - :class:`~cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey`
    """

    def __hash__(self) -> int:
        # public_numbers() hasn't got stable hash!
        # https://github.com/pyca/cryptography/issues/2143
        if isinstance(self._wrapped, ec.EllipticCurvePrivateKeyWithSerialization):
            priv = self.private_numbers()
            pub = priv.public_numbers
            return hash((self.__class__, pub.curve.name, pub.x, pub.y, priv.private_value))
        elif isinstance(self._wrapped, ec.EllipticCurvePublicKeyWithSerialization):
            pub = self.public_numbers()
            return hash((self.__class__, pub.curve.name, pub.x, pub.y))

        raise NotImplementedError()


GenericImmutableMap = TypeVar("GenericImmutableMap", bound="ImmutableMap")


class ImmutableMap(Mapping, Hashable):
    """Immutable key to value mapping with attribute access."""

    __slots__: Tuple[str, ...] = ()
    """Must be overridden in subclasses."""

    def __init__(self, **kwargs: Any) -> None:
        if set(kwargs) != set(self.__slots__):
            raise TypeError(
                "__init__() takes exactly the following arguments: {0} "
                "({1} given)".format(
                    ", ".join(self.__slots__), ", ".join(kwargs) if kwargs else "none"
                )
            )
        for slot in self.__slots__:
            object.__setattr__(self, slot, kwargs.pop(slot))

    def update(self: GenericImmutableMap, **kwargs: Any) -> GenericImmutableMap:
        """Return updated map."""
        items: Mapping[str, Any] = {**self, **kwargs}
        return type(self)(**items)

    def __getitem__(self, key: str) -> Any:
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__slots__)

    def __len__(self) -> int:
        return len(self.__slots__)

    def __hash__(self) -> int:
        return hash(tuple(getattr(self, slot) for slot in self.__slots__))

    def __setattr__(self, name: str, value: Any) -> None:
        raise AttributeError("can't set attribute")

    def __repr__(self) -> str:
        return "{0}({1})".format(
            self.__class__.__name__,
            ", ".join("{0}={1!r}".format(key, value) for key, value in self.items()),
        )


class frozendict(Mapping, Hashable):
    """Frozen dictionary."""

    __slots__ = ("_items", "_keys")

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        items: Mapping
        if kwargs and not args:
            items = dict(kwargs)
        elif len(args) == 1 and isinstance(args[0], Mapping):
            items = args[0]
        else:
            raise TypeError()
        # TODO: support generators/iterators

        object.__setattr__(self, "_items", items)
        object.__setattr__(self, "_keys", tuple(sorted(items.keys())))

    def __getitem__(self, key: str) -> Any:
        return self._items[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._keys)

    def __len__(self) -> int:
        return len(self._items)

    def _sorted_items(self) -> Tuple[Tuple[str, Any], ...]:
        return tuple((key, self[key]) for key in self._keys)

    def __hash__(self) -> int:
        return hash(self._sorted_items())

    def __getattr__(self, name: str) -> Any:
        try:
            return self._items[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name: str, value: Any) -> None:
        raise AttributeError("can't set attribute")

    def __repr__(self) -> str:
        return "frozendict({0})".format(
            ", ".join("{0}={1!r}".format(key, value) for key, value in self._sorted_items())
        )


# This class takes a similar approach to the cryptography project to deprecate attributes
# in public modules. See the _ModuleWithDeprecation class here:
# https://github.com/pyca/cryptography/blob/91105952739442a74582d3e62b3d2111365b0dc7/src/cryptography/utils.py#L129
class _UtilDeprecationModule:
    """
    Internal class delegating to a module, and displaying warnings when attributes
    related to the deprecated "abstractclassmethod" attributes in the josepy.util module.
    """

    def __init__(self, module: ModuleType) -> None:
        self.__dict__["_module"] = module

    def __getattr__(self, attr: str) -> Any:
        if attr == "abstractclassmethod":
            warnings.warn(
                "The abstractclassmethod attribute in josepy.util is deprecated and will "
                "be removed soon. Please use the built-in decorators @classmethod and "
                "@abc.abstractmethod together instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        return getattr(self._module, attr)

    def __setattr__(self, attr: str, value: Any) -> None:  # pragma: no cover
        setattr(self._module, attr, value)

    def __delattr__(self, attr: str) -> None:  # pragma: no cover
        delattr(self._module, attr)

    def __dir__(self) -> List[str]:  # pragma: no cover
        return ["_module"] + dir(self._module)


# Patching ourselves to warn about deprecation and planned removal of some elements in the module.
sys.modules[__name__] = cast(ModuleType, _UtilDeprecationModule(sys.modules[__name__]))



"""Javascript Object Signing and Encryption (JOSE).

This package is a Python implementation of the standards developed by
IETF `Javascript Object Signing and Encryption (Active WG)`_, in
particular the following RFCs:

- `JSON Web Algorithms (JWA)`_
- `JSON Web Key (JWK)`_
- `JSON Web Signature (JWS)`_

Originally developed as part of the ACME_ protocol implementation.

.. _`Javascript Object Signing and Encryption (Active WG)`:
  https://tools.ietf.org/wg/jose/

.. _`JSON Web Algorithms (JWA)`:
  https://datatracker.ietf.org/doc/draft-ietf-jose-json-web-algorithms/

.. _`JSON Web Key (JWK)`:
  https://datatracker.ietf.org/doc/draft-ietf-jose-json-web-key/

.. _`JSON Web Signature (JWS)`:
  https://datatracker.ietf.org/doc/draft-ietf-jose-json-web-signature/

.. _ACME: https://pypi.python.org/pypi/acme

"""

import sys
import warnings

# flake8: noqa
from josepy.b64 import b64decode, b64encode
from josepy.errors import (
    DeserializationError,
    Error,
    SerializationError,
    UnrecognizedTypeError,
)
from josepy.interfaces import JSONDeSerializable
from josepy.json_util import (
    Field,
    JSONObjectWithFields,
    TypedJSONObjectWithFields,
    decode_b64jose,
    decode_cert,
    decode_csr,
    decode_hex16,
    encode_b64jose,
    encode_cert,
    encode_csr,
    encode_hex16,
    field,
)
from josepy.jwa import (
    ES256,
    ES384,
    ES512,
    HS256,
    HS384,
    HS512,
    PS256,
    PS384,
    PS512,
    RS256,
    RS384,
    RS512,
    JWASignature,
)
from josepy.jwk import JWK, JWKEC, JWKRSA, JWKOct
from josepy.jws import JWS, Header, Signature
from josepy.util import (
    ComparableECKey,
    ComparableKey,
    ComparableRSAKey,
    ImmutableMap,
)

if sys.version_info[:2] == (3, 9):
    warnings.warn(
        "Python 3.9 support will be dropped in the next scheduled release of "
        "josepy. Please upgrade your Python version.",
        DeprecationWarning,
    )



"""JSON Web Key."""

import abc
import json
import logging
import math
from typing import (
    Any,
    Callable,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import cryptography.exceptions
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec, rsa

import josepy.util
from josepy import errors, json_util, util

logger = logging.getLogger(__name__)


class JWK(json_util.TypedJSONObjectWithFields, metaclass=abc.ABCMeta):
    """JSON Web Key."""

    type_field_name = "kty"
    TYPES: Dict[str, Type["JWK"]] = {}
    cryptography_key_types: Tuple[Type[Any], ...] = ()
    """Subclasses should override."""

    required: Sequence[str] = NotImplemented
    """Required members of public key's representation as defined by JWK/JWA."""

    _thumbprint_json_dumps_params: Dict[str, Union[Optional[int], Sequence[str], bool]] = {
        # "no whitespace or line breaks before or after any syntactic
        # elements"
        "indent": None,
        "separators": (",", ":"),
        # "members ordered lexicographically by the Unicode [UNICODE]
        # code points of the member names"
        "sort_keys": True,
    }
    key: Any

    def thumbprint(
        self, hash_function: Callable[[], hashes.HashAlgorithm] = hashes.SHA256
    ) -> bytes:
        """Compute JWK Thumbprint.

        https://tools.ietf.org/html/rfc7638

        :returns: bytes

        """
        digest = hashes.Hash(hash_function(), backend=default_backend())
        digest.update(
            json.dumps(
                {k: v for k, v in self.to_json().items() if k in self.required},
                **self._thumbprint_json_dumps_params,  # type: ignore[arg-type]
            ).encode()
        )
        return digest.finalize()

    @abc.abstractmethod
    def public_key(self) -> "JWK":  # pragma: no cover
        """Generate JWK with public key.

        For symmetric cryptosystems, this would return ``self``.

        """
        raise NotImplementedError()

    @classmethod
    def _load_cryptography_key(
        cls, data: bytes, password: Optional[bytes] = None, backend: Optional[Any] = None
    ) -> Any:
        backend = default_backend() if backend is None else backend
        exceptions = {}

        # private key?
        loader_private: Any
        for loader_private in (
            serialization.load_pem_private_key,
            serialization.load_der_private_key,
        ):
            try:
                return loader_private(data, password, backend)
            except (ValueError, TypeError, cryptography.exceptions.UnsupportedAlgorithm) as error:
                exceptions[str(loader_private)] = error

        # public key?
        loader_public: Any
        for loader_public in (serialization.load_pem_public_key, serialization.load_der_public_key):
            try:
                return loader_public(data, backend)
            except (ValueError, cryptography.exceptions.UnsupportedAlgorithm) as error:
                exceptions[str(loader_public)] = error

        # no luck
        raise errors.Error("Unable to deserialize key: {0}".format(exceptions))

    @classmethod
    def load(
        cls, data: bytes, password: Optional[bytes] = None, backend: Optional[Any] = None
    ) -> "JWK":
        """Load serialized key as JWK.

        :param str data: Public or private key serialized as PEM or DER.
        :param str password: Optional password.
        :param backend: A `.PEMSerializationBackend` and
            `.DERSerializationBackend` provider.

        :raises errors.Error: if unable to deserialize, or unsupported
            JWK algorithm

        :returns: JWK of an appropriate type.
        :rtype: `JWK`

        """
        try:
            key = cls._load_cryptography_key(data, password, backend)
        except errors.Error as error:
            logger.debug("Loading symmetric key, asymmetric failed: %s", error)
            return JWKOct(key=data)

        if cls.typ is not NotImplemented and not isinstance(key, cls.cryptography_key_types):
            raise errors.Error(
                "Unable to deserialize {0} into {1}".format(key.__class__, cls.__class__)
            )
        for jwk_cls in cls.TYPES.values():
            if isinstance(key, jwk_cls.cryptography_key_types):
                return jwk_cls(key=key)
        raise errors.Error("Unsupported algorithm: {0}".format(key.__class__))


@JWK.register
class JWKOct(JWK):
    """Symmetric JWK."""

    typ = "oct"
    __slots__ = ("key",)
    required = ("k", JWK.type_field_name)
    key: bytes

    def fields_to_partial_json(self) -> Dict[str, str]:
        # TODO: An "alg" member SHOULD also be present to identify the
        # algorithm intended to be used with the key, unless the
        # application uses another means or convention to determine
        # the algorithm used.
        return {"k": json_util.encode_b64jose(self.key)}

    @classmethod
    def fields_from_json(cls, jobj: Mapping[str, Any]) -> "JWKOct":
        return cls(key=json_util.decode_b64jose(jobj["k"]))

    def public_key(self) -> "JWKOct":
        return self


@JWK.register
class JWKRSA(JWK):
    """RSA JWK.

    :ivar key: :class:`~cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey`
        or :class:`~cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey` wrapped
        in :class:`~josepy.util.ComparableRSAKey`

    """

    typ = "RSA"
    cryptography_key_types = (rsa.RSAPublicKey, rsa.RSAPrivateKey)
    __slots__ = ("key",)
    required = ("e", JWK.type_field_name, "n")
    key: josepy.util.ComparableRSAKey

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if "key" in kwargs and not isinstance(kwargs["key"], util.ComparableRSAKey):
            kwargs["key"] = util.ComparableRSAKey(kwargs["key"])
        super().__init__(*args, **kwargs)

    @classmethod
    def _encode_param(cls, data: int) -> str:
        """Encode Base64urlUInt.
        :type data: long
        :rtype: unicode
        """
        length = max(data.bit_length(), 8)  # decoding 0
        length = math.ceil(length / 8)
        return json_util.encode_b64jose(data.to_bytes(byteorder="big", length=length))

    @classmethod
    def _decode_param(cls, data: str) -> int:
        """Decode Base64urlUInt."""
        try:
            binary = json_util.decode_b64jose(data)
            if not binary:
                raise errors.DeserializationError()
            return int.from_bytes(binary, byteorder="big")
        except ValueError:  # invalid literal for long() with base 16
            raise errors.DeserializationError()

    def public_key(self) -> "JWKRSA":
        return type(self)(key=self.key.public_key())

    @classmethod
    def fields_from_json(cls, jobj: Mapping[str, Any]) -> "JWKRSA":
        n, e = (cls._decode_param(jobj[x]) for x in ("n", "e"))
        public_numbers = rsa.RSAPublicNumbers(e=e, n=n)

        # public key
        if "d" not in jobj:
            return cls(key=public_numbers.public_key(default_backend()))

        # private key
        d = cls._decode_param(jobj["d"])
        if (
            "p" in jobj
            or "q" in jobj
            or "dp" in jobj
            or "dq" in jobj
            or "qi" in jobj
            or "oth" in jobj
        ):
            # "If the producer includes any of the other private
            # key parameters, then all of the others MUST be
            # present, with the exception of "oth", which MUST
            # only be present when more than two prime factors
            # were used."
            (
                p,
                q,
                dp,
                dq,
                qi,
            ) = all_params = tuple(jobj.get(x) for x in ("p", "q", "dp", "dq", "qi"))
            if tuple(param for param in all_params if param is None):
                raise errors.Error("Some private parameters are missing: {0}".format(all_params))
            p, q, dp, dq, qi = tuple(cls._decode_param(str(x)) for x in all_params)

            # TODO: check for oth
        else:
            # cryptography>=0.8
            p, q = rsa.rsa_recover_prime_factors(n, e, d)
            dp = rsa.rsa_crt_dmp1(d, p)
            dq = rsa.rsa_crt_dmq1(d, q)
            qi = rsa.rsa_crt_iqmp(p, q)

        key = rsa.RSAPrivateNumbers(p, q, d, dp, dq, qi, public_numbers).private_key(
            default_backend()
        )

        return cls(key=key)

    def fields_to_partial_json(self) -> Dict[str, Any]:
        if isinstance(self.key._wrapped, rsa.RSAPublicKey):
            numbers = self.key.public_numbers()
            params = {
                "n": numbers.n,
                "e": numbers.e,
            }
        else:  # rsa.RSAPrivateKey
            private = self.key.private_numbers()
            public = self.key.public_key().public_numbers()
            params = {
                "n": public.n,
                "e": public.e,
                "d": private.d,
                "p": private.p,
                "q": private.q,
                "dp": private.dmp1,
                "dq": private.dmq1,
                "qi": private.iqmp,
            }
        return {key: self._encode_param(value) for key, value in params.items()}


@JWK.register
class JWKEC(JWK):
    """EC JWK.

    :ivar key: :class:`~cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey`
        or :class:`~cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey` wrapped
        in :class:`~josepy.util.ComparableECKey`

    """

    typ = "EC"
    __slots__ = ("key",)
    cryptography_key_types = (ec.EllipticCurvePublicKey, ec.EllipticCurvePrivateKey)
    required = ("crv", JWK.type_field_name, "x", "y")
    key: josepy.util.ComparableECKey

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if "key" in kwargs and not isinstance(kwargs["key"], util.ComparableECKey):
            kwargs["key"] = util.ComparableECKey(kwargs["key"])
        super().__init__(*args, **kwargs)

    @classmethod
    def _encode_param(cls, data: int, length: int) -> str:
        """Encode Base64urlUInt.
        :type data: long
        :type key_size: long
        :rtype: unicode
        """
        return json_util.encode_b64jose(data.to_bytes(byteorder="big", length=length))

    @classmethod
    def _decode_param(cls, data: str, name: str, valid_length: int) -> int:
        """Decode Base64urlUInt."""
        try:
            binary = json_util.decode_b64jose(data)
            if len(binary) != valid_length:
                raise errors.DeserializationError(
                    f'Expected parameter "{name}" to be {valid_length} bytes '
                    f"after base64-decoding; got {len(binary)} bytes instead"
                )
            return int.from_bytes(binary, byteorder="big")
        except ValueError:  # invalid literal for long() with base 16
            raise errors.DeserializationError()

    @classmethod
    def _curve_name_to_crv(cls, curve_name: str) -> str:
        if curve_name == "secp256r1":
            return "P-256"
        if curve_name == "secp384r1":
            return "P-384"
        if curve_name == "secp521r1":
            return "P-521"
        raise errors.SerializationError()

    @classmethod
    def _crv_to_curve(cls, crv: str) -> ec.EllipticCurve:
        # crv is case-sensitive
        if crv == "P-256":
            return ec.SECP256R1()
        if crv == "P-384":
            return ec.SECP384R1()
        if crv == "P-521":
            return ec.SECP521R1()
        raise errors.DeserializationError()

    @classmethod
    def expected_length_for_curve(cls, curve: ec.EllipticCurve) -> int:
        if isinstance(curve, ec.SECP256R1):
            return 32
        elif isinstance(curve, ec.SECP384R1):
            return 48
        elif isinstance(curve, ec.SECP521R1):
            return 66
        raise ValueError(f"Unexpected curve: {curve}")

    def fields_to_partial_json(self) -> Dict[str, Any]:
        params = {}
        if isinstance(self.key._wrapped, ec.EllipticCurvePublicKey):
            public = self.key.public_numbers()
        elif isinstance(self.key._wrapped, ec.EllipticCurvePrivateKey):
            private = self.key.private_numbers()
            public = self.key.public_key().public_numbers()
            params["d"] = private.private_value
        else:
            raise errors.SerializationError(
                "Supplied key is neither of type EllipticCurvePublicKey "
                "nor EllipticCurvePrivateKey"
            )
        params["x"] = public.x
        params["y"] = public.y
        params = {
            key: self._encode_param(value, self.expected_length_for_curve(public.curve))
            for key, value in params.items()
        }
        params["crv"] = self._curve_name_to_crv(public.curve.name)
        return params

    @classmethod
    def fields_from_json(cls, jobj: Mapping[str, Any]) -> "JWKEC":
        curve = cls._crv_to_curve(jobj["crv"])
        expected_length = cls.expected_length_for_curve(curve)
        x, y = (cls._decode_param(jobj[n], n, expected_length) for n in ("x", "y"))
        public_numbers = ec.EllipticCurvePublicNumbers(x=x, y=y, curve=curve)

        # private key
        if "d" not in jobj:
            return cls(key=public_numbers.public_key(default_backend()))

        # private key
        d = cls._decode_param(jobj["d"], "d", expected_length)
        key = ec.EllipticCurvePrivateNumbers(d, public_numbers).private_key(default_backend())
        return cls(key=key)

    def public_key(self) -> "JWKEC":
        # Unlike RSAPrivateKey, EllipticCurvePrivateKey does not contain public_key()
        if hasattr(self.key, "public_key"):
            key = self.key.public_key()
        else:
            key = self.key.public_numbers().public_key(default_backend())
        return type(self)(key=key)



"""JSON Web Signature."""

import argparse
import base64
import sys
from typing import (
    Any,
    Dict,
    FrozenSet,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    cast,
)

from cryptography import x509
from cryptography.hazmat.primitives.serialization import Encoding

import josepy
from josepy import b64, errors, json_util, jwa
from josepy import jwk as jwk_mod


class MediaType:
    """MediaType field encoder/decoder."""

    PREFIX = "application/"
    """MIME Media Type and Content Type prefix."""

    @classmethod
    def decode(cls, value: str) -> str:
        """Decoder."""
        # 4.1.10
        if "/" not in value:
            if ";" in value:
                raise errors.DeserializationError("Unexpected semi-colon")
            return cls.PREFIX + value
        return value

    @classmethod
    def encode(cls, value: str) -> str:
        """Encoder."""
        # 4.1.10
        if ";" not in value:
            assert value.startswith(cls.PREFIX)
            return value[len(cls.PREFIX) :]
        return value


class Header(json_util.JSONObjectWithFields):
    """JOSE Header.

    .. warning:: This class supports **only** Registered Header
        Parameter Names (as defined in section 4.1 of the
        protocol). If you need Public Header Parameter Names (4.2)
        or Private Header Parameter Names (4.3), you must subclass
        and override :meth:`from_json` and :meth:`to_partial_json`
        appropriately.

    .. warning:: This class does not support any extensions through
        the "crit" (Critical) Header Parameter (4.1.11) and as a
        conforming implementation, :meth:`from_json` treats its
        occurrence as an error. Please subclass if you seek for
        a different behaviour.

    :ivar x5tS256: "x5t#S256"
    :ivar str typ: MIME Media Type, inc. :const:`MediaType.PREFIX`.
    :ivar str cty: Content-Type, inc. :const:`MediaType.PREFIX`.

    """

    alg: Optional[jwa.JWASignature] = json_util.field(
        "alg", decoder=jwa.JWASignature.from_json, omitempty=True
    )
    jku: Optional[bytes] = json_util.field("jku", omitempty=True)
    jwk: Optional[jwk_mod.JWK] = json_util.field(
        "jwk", decoder=jwk_mod.JWK.from_json, omitempty=True
    )
    kid: Optional[str] = json_util.field("kid", omitempty=True)
    x5u: Optional[bytes] = json_util.field("x5u", omitempty=True)
    x5c: Tuple[x509.Certificate, ...] = json_util.field("x5c", omitempty=True, default=())
    x5t: Optional[bytes] = json_util.field("x5t", decoder=json_util.decode_b64jose, omitempty=True)
    x5tS256: Optional[bytes] = json_util.field(
        "x5t#S256", decoder=json_util.decode_b64jose, omitempty=True
    )
    typ: Optional[MediaType] = json_util.field(
        "typ", encoder=MediaType.encode, decoder=MediaType.decode, omitempty=True
    )
    cty: Optional[MediaType] = json_util.field(
        "cty", encoder=MediaType.encode, decoder=MediaType.decode, omitempty=True
    )
    crit: Tuple[Any, ...] = json_util.field("crit", omitempty=True, default=())
    _fields: Dict[str, json_util.Field]

    def not_omitted(self) -> Dict[str, json_util.Field]:
        """Fields that would not be omitted in the JSON object."""
        return {
            name: getattr(self, name)
            for name, field in self._fields.items()
            if not field.omit(getattr(self, name))
        }

    def __add__(self, other: Any) -> "Header":
        if not isinstance(other, type(self)):
            raise TypeError("Header cannot be added to: {0}".format(type(other)))

        not_omitted_self = self.not_omitted()
        not_omitted_other = other.not_omitted()

        if set(not_omitted_self).intersection(not_omitted_other):
            raise TypeError("Addition of overlapping headers not defined")

        not_omitted_self.update(not_omitted_other)
        return type(self)(**not_omitted_self)

    def find_key(self) -> josepy.JWK:
        """Find key based on header.

        .. todo:: Supports only "jwk" header parameter lookup.

        :returns: (Public) key found in the header.
        :rtype: .JWK

        :raises josepy.errors.Error: if key could not be found

        """
        if self.jwk is None:
            raise errors.Error("No key found")
        return self.jwk

    @crit.decoder  # type: ignore
    def crit(unused_value: Any) -> Any:
        raise errors.DeserializationError('"crit" is not supported, please subclass')

    # x5c does NOT use JOSE Base64 (4.1.6)

    @x5c.encoder  # type: ignore
    def x5c(value):
        """
        .. versionchanged:: 2.0.0
           The values are now `cryptography.x509.Certificate` objects.
           Previously these were `josepy.util.ComparableX509` objects, which wrapped
           `OpenSSL.crypto.X509` objects.
        """
        return [base64.b64encode(cert.public_bytes(Encoding.DER)) for cert in value]

    @x5c.decoder  # type: ignore
    def x5c(value):
        """
        .. versionchanged:: 2.0.0
           The values are now `cryptography.x509.Certificate` objects.
           Previously these were `josepy.util.ComparableX509` objects, which wrapped
           `OpenSSL.crypto.X509` objects.
        """
        try:
            return tuple(x509.load_der_x509_certificate(base64.b64decode(cert)) for cert in value)
        except ValueError as error:
            raise errors.DeserializationError(error)


class Signature(json_util.JSONObjectWithFields):
    """JWS Signature.

    :ivar combined: Combined Header (protected and unprotected,
        :class:`Header`).
    :ivar unicode protected: JWS protected header (Jose Base-64 decoded).
    :ivar header: JWS Unprotected Header (:class:`Header`).
    :ivar str signature: The signature.

    """

    header_cls = Header
    combined: Header

    __slots__ = ("combined",)
    protected: str = json_util.field("protected", omitempty=True, default="")
    header: Header = json_util.field(
        "header", omitempty=True, default=header_cls(), decoder=header_cls.from_json
    )
    signature: bytes = json_util.field(
        "signature", decoder=json_util.decode_b64jose, encoder=json_util.encode_b64jose
    )

    @protected.encoder  # type: ignore
    def protected(value: str) -> str:
        # wrong type guess (Signature, not bytes) | pylint: disable=no-member
        return json_util.encode_b64jose(value.encode("utf-8"))

    @protected.decoder  # type: ignore
    def protected(value: str) -> str:
        return json_util.decode_b64jose(value).decode("utf-8")

    def __init__(self, **kwargs: Any) -> None:
        if "combined" not in kwargs:
            kwargs = self._with_combined(kwargs)
        super().__init__(**kwargs)
        assert self.combined.alg is not None

    @classmethod
    def _with_combined(cls, kwargs: Any) -> Dict[str, Any]:
        assert "combined" not in kwargs
        header = kwargs.get("header", cls._fields["header"].default)
        protected = kwargs.get("protected", cls._fields["protected"].default)

        if protected:
            combined = header + cls.header_cls.json_loads(protected)
        else:
            combined = header

        kwargs["combined"] = combined
        return kwargs

    @classmethod
    def _msg(cls, protected: str, payload: bytes) -> bytes:
        return b64.b64encode(protected.encode("utf-8")) + b"." + b64.b64encode(payload)

    def verify(self, payload: bytes, key: Optional[josepy.JWK] = None) -> bool:
        """Verify.

        :param bytes payload: Payload to verify.
        :param JWK key: Key used for verification.

        """
        actual_key: josepy.JWK = self.combined.find_key() if key is None else key
        if not self.combined.alg:
            raise josepy.Error("Not signature algorithm defined.")
        return self.combined.alg.verify(
            key=actual_key.key, sig=self.signature, msg=self._msg(self.protected, payload)
        )

    @classmethod
    def sign(
        cls,
        payload: bytes,
        key: josepy.JWK,
        alg: josepy.JWASignature,
        include_jwk: bool = True,
        protect: FrozenSet = frozenset(),
        **kwargs: Any,
    ) -> "Signature":
        """Sign.

        :param bytes payload: Payload to sign.
        :param JWK key: Key for signature.
        :param JWASignature alg: Signature algorithm to use to sign.
        :param bool include_jwk: If True, insert the JWK inside the signature headers.
        :param FrozenSet protect: List of headers to protect.

        """
        assert isinstance(key, alg.kty)

        header_params = kwargs
        header_params["alg"] = alg
        if include_jwk:
            header_params["jwk"] = key.public_key()

        assert set(header_params).issubset(cls.header_cls._fields)
        assert protect.issubset(cls.header_cls._fields)

        protected_params = {}
        for header in protect:
            if header in header_params:
                protected_params[header] = header_params.pop(header)
        if protected_params:
            protected = cls.header_cls(**protected_params).json_dumps()
        else:
            protected = ""

        header = cls.header_cls(**header_params)
        signature = alg.sign(key.key, cls._msg(protected, payload))

        return cls(protected=protected, header=header, signature=signature)

    def fields_to_partial_json(self) -> Dict[str, Any]:
        fields = super().fields_to_partial_json()
        if not fields["header"].not_omitted():
            del fields["header"]
        return fields

    @classmethod
    def fields_from_json(cls, jobj: Mapping[str, Any]) -> Dict[str, Any]:
        fields = super().fields_from_json(jobj)
        fields_with_combined = cls._with_combined(fields)
        if "alg" not in fields_with_combined["combined"].not_omitted():
            raise errors.DeserializationError("alg not present")
        return fields_with_combined


class JWS(json_util.JSONObjectWithFields):
    """JSON Web Signature.

    :ivar str payload: JWS Payload.
    :ivar str signature: JWS Signatures.

    """

    __slots__ = ("payload", "signatures")
    payload: bytes
    signatures: List[Signature]

    signature_cls = Signature

    def verify(self, key: Optional[josepy.JWK] = None) -> bool:
        """Verify."""
        return all(sig.verify(self.payload, key) for sig in self.signatures)

    @classmethod
    def sign(cls, payload: bytes, **kwargs: Any) -> "JWS":
        """Sign."""
        return cls(payload=payload, signatures=(cls.signature_cls.sign(payload=payload, **kwargs),))

    @property
    def signature(self) -> Signature:
        """Get a singleton signature.

        :rtype: :class:`JWS.signature_cls`

        """
        assert len(self.signatures) == 1
        return self.signatures[0]

    def to_compact(self) -> bytes:
        """Compact serialization.

        :rtype: bytes

        """
        assert len(self.signatures) == 1

        assert "alg" not in self.signature.header.not_omitted()
        # ... it must be in protected

        return (
            b64.b64encode(self.signature.protected.encode("utf-8"))
            + b"."
            + b64.b64encode(self.payload)
            + b"."
            + b64.b64encode(self.signature.signature)
        )

    @classmethod
    def from_compact(cls, compact: bytes) -> "JWS":
        """Compact deserialization.

        :param bytes compact:

        """
        try:
            protected, payload, signature = compact.split(b".")
        except ValueError:
            raise errors.DeserializationError(
                "Compact JWS serialization should comprise of exactly" " 3 dot-separated components"
            )

        sig = cls.signature_cls(
            protected=b64.b64decode(protected).decode("utf-8"), signature=b64.b64decode(signature)
        )
        return cls(payload=b64.b64decode(payload), signatures=(sig,))

    def to_partial_json(self, flat: bool = True) -> Dict[str, Any]:
        assert self.signatures
        payload = json_util.encode_b64jose(self.payload)

        if flat and len(self.signatures) == 1:
            ret = self.signatures[0].to_partial_json()
            ret["payload"] = payload
            return ret
        else:
            return {
                "payload": payload,
                "signatures": self.signatures,
            }

    @classmethod
    def from_json(cls, jobj: Mapping[str, Any]) -> "JWS":
        if "signature" in jobj and "signatures" in jobj:
            raise errors.DeserializationError("Flat mixed with non-flat")
        elif "signature" in jobj:  # flat
            filtered = {key: value for key, value in jobj.items() if key != "payload"}
            return cls(
                payload=json_util.decode_b64jose(jobj["payload"]),
                signatures=(cls.signature_cls.from_json(filtered),),
            )
        else:
            return cls(
                payload=json_util.decode_b64jose(jobj["payload"]),
                signatures=tuple(cls.signature_cls.from_json(sig) for sig in jobj["signatures"]),
            )


class CLI:
    """JWS CLI."""

    @classmethod
    def sign(cls, args: argparse.Namespace) -> None:
        """Sign."""
        with open(args.key, "rb") as fp:
            key = args.alg.kty.load(fp.read())
        if args.protect is None:
            args.protect = []
        if args.compact:
            args.protect.append("alg")

        sig = JWS.sign(
            payload=sys.stdin.read().encode(), key=key, alg=args.alg, protect=set(args.protect)
        )

        if args.compact:
            print(sig.to_compact().decode("utf-8"))
        else:  # JSON
            print(sig.json_dumps_pretty())

    @classmethod
    def verify(cls, args: argparse.Namespace) -> bool:
        """Verify."""
        if args.compact:
            sig = JWS.from_compact(sys.stdin.read().encode())
        else:  # JSON
            try:
                sig = cast(JWS, JWS.json_loads(sys.stdin.read()))
            except errors.Error as error:
                print(error)
                return False

        if args.key is not None:
            assert args.kty is not None
            with open(args.key, "rb") as fp:
                key = args.kty.load(fp.read()).public_key()
        else:
            key = None

        sys.stdout.write(sig.payload.decode())
        return not sig.verify(key=key)

    @classmethod
    def _alg_type(cls, arg: Any) -> jwa.JWASignature:
        return jwa.JWASignature.from_json(arg)

    @classmethod
    def _header_type(cls, arg: Any) -> Any:
        assert arg in Signature.header_cls._fields
        return arg

    @classmethod
    def _kty_type(cls, arg: Any) -> Type[jwk_mod.JWK]:
        assert arg in jwk_mod.JWK.TYPES
        return jwk_mod.JWK.TYPES[arg]

    @classmethod
    def run(cls, args: Optional[List[str]] = None) -> Optional[bool]:
        """Parse arguments and sign/verify."""
        if args is None:
            args = sys.argv[1:]
        parser = argparse.ArgumentParser()
        parser.add_argument("--compact", action="store_true")

        subparsers = parser.add_subparsers()
        parser_sign = subparsers.add_parser("sign")
        parser_sign.set_defaults(func=cls.sign)
        parser_sign.add_argument("-k", "--key", required=True)
        parser_sign.add_argument("-a", "--alg", type=cls._alg_type, default=jwa.RS256)
        parser_sign.add_argument("-p", "--protect", action="append", type=cls._header_type)

        parser_verify = subparsers.add_parser("verify")
        parser_verify.set_defaults(func=cls.verify)
        parser_verify.add_argument("-k", "--key", required=False)
        parser_verify.add_argument("--kty", type=cls._kty_type, required=False)

        parsed = parser.parse_args(args)
        return parsed.func(parsed)


if __name__ == "__main__":
    exit(CLI.run())  # pragma: no cover
